use crate::server::ServerProcess;
use clap::ValueEnum;
use std::fmt;
use std::fs::File;
use std::io::Write;
use std::path::{Path, PathBuf};
use std::process::{Command, Stdio};

#[derive(Copy, Clone, PartialEq, Eq, PartialOrd, Ord, ValueEnum)]
pub enum ServerType {
    Paper,
    Forge,
    Vanilla,
}

impl fmt::Display for ServerType {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let s = match self {
            ServerType::Paper => "PaperMC",
            ServerType::Forge => "Forge",
            ServerType::Vanilla => "Vanilla",
        };

        write!(f, "{}", s)
    }
}

pub struct ServerCommand {
    type_: ServerType,
    version: String,
    java: String,
    jar: PathBuf,
    config_dir: PathBuf,
    world_dir: PathBuf,
    backup_dir: PathBuf,
    xms: u64,
    xmx: u64,
    max_backups: u64,
}

impl ServerCommand {
    pub fn new(type_: ServerType, version: &str) -> Self {
        ServerCommand {
            type_,
            version: String::from(version),
            java: String::from("java"),
            jar: PathBuf::from("server.jar"),
            config_dir: PathBuf::from("config"),
            world_dir: PathBuf::from("worlds"),
            backup_dir: PathBuf::from("backups"),
            xms: 1024,
            xmx: 2048,
            max_backups: 7,
        }
    }

    pub fn java(mut self, java: &str) -> Self {
        self.java = String::from(java);

        self
    }

    pub fn jar<T: AsRef<Path>>(mut self, path: T) -> Self {
        self.jar = PathBuf::from(path.as_ref());
        self
    }

    pub fn config<T: AsRef<Path>>(mut self, path: T) -> Self {
        self.config_dir = PathBuf::from(path.as_ref());
        self
    }

    pub fn world<T: AsRef<Path>>(mut self, path: T) -> Self {
        self.world_dir = PathBuf::from(path.as_ref());
        self
    }

    pub fn backup<T: AsRef<Path>>(mut self, path: T) -> Self {
        self.backup_dir = PathBuf::from(path.as_ref());

        self
    }

    pub fn xms(mut self, v: u64) -> Self {
        self.xms = v;
        self
    }

    pub fn xmx(mut self, v: u64) -> Self {
        self.xmx = v;
        self
    }

    pub fn max_backups(mut self, v: u64) -> Self {
        self.max_backups = v;
        self
    }

    fn accept_eula(&self) -> std::io::Result<()> {
        let mut eula_path = self.config_dir.clone();
        eula_path.push("eula.txt");
        let mut eula_file = File::create(eula_path)?;
        eula_file.write_all(b"eula=true")?;

        Ok(())
    }

    pub fn spawn(self) -> std::io::Result<ServerProcess> {
        // To avoid any issues, we use absolute paths for everything when spawning the process
        let jar = self.jar.canonicalize()?;
        let config_dir = self.config_dir.canonicalize()?;
        let world_dir = self.world_dir.canonicalize()?;
        let backup_dir = self.backup_dir.canonicalize()?;

        self.accept_eula()?;

        let child = Command::new(&self.java)
            .current_dir(&config_dir)
            .arg("-jar")
            .arg(&jar)
            .arg("--universe")
            .arg(&world_dir)
            .arg("--nogui")
            .stdin(Stdio::piped())
            .spawn()?;

        Ok(ServerProcess::new(
            self.type_,
            self.version,
            config_dir,
            world_dir,
            backup_dir,
            self.max_backups,
            child,
        ))
    }
}
